import datetime
import uuid
from typing import Dict, List, Optional

from O365 import Account as BaseAccount
from O365 import FileSystemTokenBackend, MSOffice365Protocol
from O365.calendar import Calendar, Event
from O365.connection import Connection as BaseConnection
from O365.mailbox import MailBox
from O365.utils import Query

from . import errors


class Connection(BaseConnection):
    def set_proxy(self, proxy_server, proxy_port, proxy_username, proxy_password):
        """Sets a proxy on the Session"""
        # We want to set an 'https' proxy with a 'http' endpoint in order to do
        # tunneling over HTTP CONNECT.
        # self.proxy is used as the 'proxies' keyword argument to requests,
        # so if proxy_server is a dict we just pass it on.
        if isinstance(proxy_server, dict):
            self.proxy = proxy_server
        else:
            super().set_proxy(proxy_server, proxy_port, proxy_username, proxy_password)


class Account(BaseAccount):
    connection_constructor = Connection


class ManualQuery:
    """
    A very basic, manual version of O365.utils.Query which
    only provides the as_params method.
    """

    def __init__(self, params=None):
        self.params = params

    def as_params(self) -> dict:
        """Returns the filters, orders, select, expands and search as query parameters

        :rtype: dict
        """
        return self.params or {}


class EventWithExtendedProperties(Event):
    """
    Extends O365.calendar.Event with support for singleValueExtendedProperties.

    To fetch events with this field populated, an $expand query parameter
    filtering for the exact extended property ID must be included.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        cloud_data = kwargs.get(self._cloud_data_key, {})
        self.__single_value_extended_properties = cloud_data.get(
            self._cc("singleValueExtendedProperties"), []
        )

    @property
    def single_value_extended_properties(self):
        return self.__single_value_extended_properties

    @single_value_extended_properties.setter
    def single_value_extended_properties(self, value):
        if isinstance(value, list):
            self.__single_value_extended_properties = value
        else:
            raise ValueError("categories must be a list")
        self._track_changes.add(self._cc("singleValueExtendedProperties"))

    def to_api_data(self, restrict_keys=None):
        data = super().to_api_data(restrict_keys=restrict_keys)
        data[
            self._cc("singleValueExtendedProperties")
        ] = self.single_value_extended_properties
        return data


class GraphClient:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        tenant_id: str,
        identity_property_name: str,
        token_path: Optional[str] = None,
        proxies: Optional[Dict[str, str]] = None,
        timezone: Optional[str] = None,
    ) -> None:
        assert client_id
        assert client_secret
        assert tenant_id
        self.credentials = (client_id, client_secret)

        assert isinstance(identity_property_name, str) and identity_property_name
        self.identity_property_name = identity_property_name

        token_backend = FileSystemTokenBackend(token_path=token_path)
        acc_kwargs = {
            "credentials": self.credentials,
            "auth_flow_type": "credentials",
            "tenant_id": tenant_id,
            "token_backend": token_backend,
            "proxy_server": proxies,
        }
        if timezone:
            acc_kwargs["timezone"] = timezone
        self.account = Account(**acc_kwargs)
        self.authenticated = self.account.authenticate()
        if self.authenticated is not True:
            raise errors.AccessDenied("initial authentication")

    def get_user_calendar(self, address: str) -> Calendar:
        """
        Fetch the parent object needed to operate on a users calendar.

        For python-o365 this is a configured O365.calendar.Calendar object.
        """
        schedule = self.account.schedule(resource=address)
        calendar = schedule.get_default_calendar()
        calendar.event_constructor = EventWithExtendedProperties
        return calendar

    def __prepare_html(self, body):
        return "<html><body>{body}</body></html>".format(body=body)

    @property
    def extended_property_id(self) -> str:
        # this is referred to as the 'id' on
        # https://docs.microsoft.com/en-us/graph/api/singlevaluelegacyextendedproperty-get?
        return "String {{7ee114a2-0646-5bdf-b578-f0bceb362b2d}} Name {}".format(
            self.identity_property_name
        )

    def _make_extended_property(self, value) -> Dict[str, str]:
        assert isinstance(value, str)
        return {
            "id": self.extended_property_id,
            "value": value,
        }

    def _filter_extended_property(self, identifier: str) -> str:
        assert identifier and "-" in identifier
        return "singleValueExtendedProperties/Any(ep: ep/id eq '{extended_property_id}' and ep/value eq '{value}')".format(
            extended_property_id=self.extended_property_id, value=identifier
        )

    def find_item(self, calendar: Calendar, identificator: str) -> Optional[Event]:
        """
        Finds a single event matching the extended property configured for use.
        """
        expander = "singleValueExtendedProperties($filter=id eq '{extended_property_id}')".format(
            extended_property_id=self.extended_property_id
        )
        query = ManualQuery(
            {
                "$filter": self._filter_extended_property(identificator),
                "$expand": expander,
            }
        )
        items = list(calendar.get_events(include_recurring=False, query=query))
        assert len(items) <= 1
        if not items:
            return None
        item = items[0]
        properties = item.single_value_extended_properties
        assert len(properties) == 1
        assert properties[0]["id"] == self.extended_property_id
        assert properties[0]["value"] == identificator
        return item

    def get_all_items(self, user_calendar: Calendar) -> List[Event]:
        expander = "singleValueExtendedProperties($filter=id eq '{extended_property_id}')".format(
            extended_property_id=self.extended_property_id
        )
        query_filter = "singleValueExtendedProperties/Any(ep: ep/id eq '{extended_property_id}' and ep/value ne null)".format(
            extended_property_id=self.extended_property_id
        )
        query = ManualQuery(
            {
                "$filter": query_filter,
                "$expand": expander,
            }
        )
        items = list(
            user_calendar.get_events(include_recurring=False, limit=None, query=query)
        )
        if not items:
            return []
        for item in items:
            assert hasattr(item, "single_value_extended_properties")
            properties = item.single_value_extended_properties
            assert len(properties) == 1
            assert properties[0]["id"] == self.extended_property_id
            assert isinstance(properties[0]["value"], str)
            assert "-" in properties[0]["value"]
        return items

    def get_identificator_from_event(self, event: Event) -> str:
        assert hasattr(event, "single_value_extended_properties")
        properties = event.single_value_extended_properties
        assert len(properties) == 1
        assert properties[0]["id"] == self.extended_property_id
        assert isinstance(properties[0]["value"], str)
        assert "-" in properties[0]["value"]
        return properties[0]["value"]

    def create_item(
        self,
        user_calendar: Calendar,
        identificator: str,
        subject: str,
        body: str,
        start: datetime.datetime,
        end: datetime.datetime,
        location: Optional[str] = None,
        cancelled: bool = False,
    ):
        if not start.tzinfo or not end.tzinfo:
            raise errors.AwareDateTimeRequired()

        event = EventWithExtendedProperties(parent=user_calendar)

        event.subject = subject
        event.body = self.__prepare_html(body)
        event.start = start
        event.end = end
        event.single_value_extended_properties = [
            self._make_extended_property(identificator)
        ]

        if location:
            event.location = location
        if cancelled:
            # TODO: setting this one is unsupported as of November 2020
            # event.is_cancelled = True if cancelled else False
            event.show_as = "free" if cancelled else "busy"

        success = event.save()
        return event if success else None

    def update_item(
        self,
        user_calendar: Calendar,
        identificator: str,
        subject: str,
        body: str,
        start: datetime.datetime,
        end: datetime.datetime,
        location: Optional[str] = None,
        cancelled: bool = False,
    ):
        if not start.tzinfo or not end.tzinfo:
            raise errors.AwareDateTimeRequired()

        event = self.find_item(user_calendar, identificator)
        if event is None:
            return False

        event.subject = subject
        event.location = location
        event.body = self.__prepare_html(body)
        event.start = start
        event.end = end
        # TODO: changing this one is unsupported as of November 2020
        # event.is_cancelled = True if cancelled else False
        event.show_as = "free" if cancelled else "busy"

        success = event.save()
        return event if success else None

    def delete_item(
        self, user_calendar: Calendar, identificator: str
    ) -> Optional[bool]:
        event = self.find_item(user_calendar, identificator)
        if not event:
            return None
        success = event.delete()
        assert isinstance(success, bool)
        return success
